"""
Draw a random number from a beta dirstribution
"""

__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

from collections import OrderedDict
import numpy as np


def is_base_string(s):
    return isinstance(s, (str, bytes))

def is_unicode_string(s):
    return isinstance(s, str)

def is_byte_string(s):
    return isinstance(s, bytes)

def join_dict_list(list_of_dicts):
    """make a dictionary of lists from a list of dictionaries"""
    rtn = OrderedDict()
    for d in list_of_dicts:
        for k, v in d.items():
            if k in rtn:
                rtn[k].append(v)
            else:
                rtn[k] = [v]
    return rtn


def dict_to_csv(dictionary, variable_names=False, dict_of_lists=False):
    d = OrderedDict(dictionary.items())
    rtn = ""
    if variable_names:
        rtn += ",".join(d.keys()) + "\n"

    if dict_of_lists:
        prop_np = np.array(list(d.values())).T  # list is requires in PY3
        for x in prop_np:
            rtn += ", ".join(map(lambda s: str(s), x)) + "\n"
    else:
        rtn += ",".join(map(lambda s: str(s), d.values())) + "\n"

    return rtn

def numpy_vector(x):
    """helper function:
    make an numpy vector from any element (list, arrays, and single data (str, numeric))
    """

    x = np.array(x)
    if x.ndim == 1:
        return x
    elif x.ndim == 0:
        return x.reshape(1)  # if one element only, make a array with one element
    else:
        return x.flatten()

def numpy_array_2d(two_d_data):
    """ensures well shaped to 2d numpy array"""
    rtn = np.array(two_d_data)
    if rtn.ndim == 1 and len(rtn) == 2:
        rtn = rtn.reshape((1, 2))
    if rtn.ndim != 2:
        raise ValueError("Bad shaped data: xy must be pair of xy-values or a list of xy-values")
    return rtn

def numpy_round2(array, decimals, int_type=np.int32):
    """rounds and changes to int type if decimals == 0"""
    array = np.round(array, decimals=decimals)
    if decimals == 0:
        return array.astype(int_type)
    else:
        return array

def is_all_larger(vector, standard=0):
    return sum(map(lambda x: x > standard, vector))==len(vector)

def is_all_smaller(vector, standard=0):
    return sum(map(lambda x: x < standard, vector))==len(vector)

def is_all_equal(vector):
    # returns true if all elements are equal
    return sum(map(lambda x: x==vector[0], vector))==len(vector)

def dict_to_text(the_dict, col_a = 22, col_b = 14,
                 spacing_char=" "):
    rtn = None
    for k, v in the_dict.items():
        if rtn is None:
            key_str = "- " + k
            rtn = ""
        else:
            key_str = "  " + k

        value = "{}\n".format(v)
        len_col_b = col_b - len(value)
        if len_col_b<2:
            len_col_b = 2
        rtn += key_str + (spacing_char * (col_a - len(key_str))) + \
                         (" " * len_col_b) + value
    return rtn.rstrip()




__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import copy
from random import shuffle

from .._lib.exceptions import NoSolutionError
from .. import arrays
from ._size_distribution import SizeDistribution


def create(reference_array,
           size_distribution,
           n_objects,
           attributes = None,
           allow_overlapping = False,
           occupied_space = None):
    """occupied_space is a dot array (used for multicolour dot array (join after)

    attribute is an array, arrays are assigned randomly.

    """
    arrays._check_base_array(reference_array)
    if not isinstance(size_distribution, SizeDistribution):
        raise RuntimeError("Size distribution has to be of type SizeDistribution, but not {}".format(
                        type(size_distribution).__name__))

    if size_distribution.diameter is not None:
        # DotArray
        rtn = arrays.DotArray(target_area_radius=reference_array.target_area_radius,
                       min_dist_between=reference_array.min_dist_between,
                       min_dist_area_boarder=reference_array.min_dist_area_boarder)

        for dot in size_distribution.sample(n=n_objects):
            try:
                dot = rtn.get_random_free_position(ref_object=dot,
                                                   occupied_space=occupied_space,
                                                   allow_overlapping=allow_overlapping)
            except NoSolutionError as e:
                raise NoSolutionError("Can't find a solution for {} items in this array".format(n_objects))
            rtn.add([dot])

    else:
        # RectArray
        rtn = arrays.RectangleArray(target_area_radius=reference_array.target_area_radius,
                             min_dist_between=reference_array.min_dist_between,
                             min_dist_area_boarder=reference_array.min_dist_area_boarder)

        for rect in size_distribution.sample(n=n_objects):
            try:
                rect = rtn.get_random_free_position(ref_object=rect,
                                                    occupied_space=occupied_space,
                                                    allow_overlapping=allow_overlapping)
            except NoSolutionError:
                raise NoSolutionError("Can't find a solution for {} items in this array".format(n_objects))

            rtn.add([rect])

    # attribute assignment
    if isinstance(attributes, (tuple, list)):
        att = []
        while len(att) < n_objects:
            tmp = copy.copy(attributes)
            shuffle(tmp)
            att.extend(tmp)
        shuffle(att)
        rtn.set_attributes(att[:n_objects])
    else:
        rtn.set_attributes(attributes)

    return rtn


def create_incremental(reference_array,
                       size_distribution,
                       n_objects,
                       attributes = None,
                       allow_overlapping = False):

    previous = None
    for n in range(n_objects):
        current = create(reference_array=reference_array,
                     size_distribution=size_distribution,
                     n_objects=1,
                     attributes=attributes,
                     allow_overlapping=allow_overlapping,
                     occupied_space=previous)
        if previous is not None:
            current.join(previous)
        previous = current
        yield current




__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import numpy as _np
import svgwrite as _svg
from . import _colour
from .._lib.geometry import cartesian2image_coordinates as _c2i_coord
from ..  import arrays as _arrays
from .. import shapes as _shapes

from ._colour import ImageColours # make available
# FIXME can't handle pictures, throw exception

def create(object_array, colours=None, filename="noname.svg"):
    _arrays._check_object_array(object_array)
    if colours is None:
        colours = _colour.ImageColours()
    if not isinstance(colours, _colour.ImageColours):
        raise TypeError("Colours must be of type image.ImageColours")

    image_size = _np.ceil(object_array.target_area_radius) * 2
    px = "{}px".format(image_size)
    svgdraw = _svg.Drawing(size = (px, px), filename=filename)

    if colours.target_area.colour is not None:
        svgdraw.add(svgdraw.circle(center=_c2i_coord(_np.zeros(2), image_size),
                                   r= object_array.target_area_radius,
                                   # stroke_width="0", stroke="black",
                                   fill=colours.target_area.colour))

    if object_array.properties.numerosity > 0:
        image_coord = _c2i_coord(object_array.xy, image_size)
        if isinstance(object_array, _arrays.DotArray):
            # draw dots
            for xy, d, att in zip(image_coord, object_array.diameters,
                                  object_array.attributes):
                obj = _shapes.Dot(xy=xy, diameter=d)
                obj.attribute = _colour.Colour(att,
                                            colours.default_object_colour)
                _draw_shape(svgdraw, obj, opacity=colours.opacity_object)

        elif isinstance(object_array, _arrays.RectangleArray):
            # draw rectangle
            for xy, size, att in zip(image_coord,
                                     object_array.sizes,
                                     object_array.attributes):
                obj = _shapes.Rectangle(xy=xy, size=size)
                obj.attribute = _colour.Colour(att,
                                            colours.default_object_colour)
                _draw_shape(svgdraw, obj, opacity=colours.opacity_object)


        # draw convex hulls
        if colours.field_area_positions.colour is not None and \
                object_array.properties.field_area_positions > 0:
            _draw_convex_hull(svgdraw=svgdraw,
                              points=_c2i_coord(
                          object_array.properties.convex_hull_positions.xy, image_size),
                              convex_hull_colour=colours.field_area_positions.colour,
                              opacity=colours.opacity_guides)
        if colours.field_area.colour is not None and \
                object_array.properties.field_area > 0:
            _draw_convex_hull(svgdraw=svgdraw,
                              points=_c2i_coord(
                          object_array.properties.convex_hull.xy,
                          image_size),
                              convex_hull_colour=colours.field_area.colour,
                              opacity=colours.opacity_guides)
        #  and center of mass
        if colours.center_of_field_area.colour is not None:
            obj = _shapes.Dot(xy=_c2i_coord(object_array.center_of_field_area(), image_size),
                              diameter=10,
                              attribute=colours.center_of_field_area.colour)
            _draw_shape(svgdraw, obj, opacity=colours.opacity_guides)
        if colours.center_of_mass.colour is not None:
            obj = _shapes.Dot(xy=_c2i_coord(object_array.center_of_mass(), image_size),
                              diameter=10,
                              attribute=colours.center_of_mass.colour)
            _draw_shape(svgdraw, obj, opacity=colours.opacity_guides)

    return svgdraw


def _draw_shape(svgdraw, shape, opacity=1):
    # draw object
    assert isinstance(shape, (_shapes.Dot, _shapes.Rectangle))

    colour = _colour.Colour(shape.attribute)

    if isinstance(shape, _shapes.Dot):
        r = shape.diameter / 2
        svgdraw.add(svgdraw.circle(center=shape.xy,
                                   r=shape.diameter/2,
                                   # stroke_width="0", stroke="black",
                                   fill=colour.colour,
                                   opacity=opacity))

    elif isinstance(shape, _shapes.Rectangle):
        svgdraw.add(svgdraw.rect(insert=(shape.left, shape.bottom),
                                 size=shape.size,
                                 fill=colour.colour,
                                 opacity=opacity))
    else:
        raise NotImplementedError("Shape {} NOT YET IMPLEMENTED".format(type(shape)))


def _draw_convex_hull(svgdraw, points, convex_hull_colour, opacity):
    # plot convey hull

    last = None
    for p in _np.append(points, [points[0]], axis=0):
        if last is not None:
            l = svgdraw.line(start=last, end=p).stroke(
                width=1, color=convex_hull_colour, opacity=opacity)
            svgdraw.add(l)
        last = p




__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import numpy as _np
from matplotlib import pyplot as _plt
from . import _colour
from .. import arrays as _arrays
from .. import shapes as _shapes

from ._colour import ImageColours # make available
# FIXME can't handle pictures
def create(object_array, colours=None, dpi=100):
    """create a matplotlib figure"""

    _arrays._check_object_array(object_array)
    if colours is None:
        colours = _colour.ImageColours()
    if not isinstance(colours, _colour.ImageColours):
        raise TypeError("Colours must be of type image.ImageColours")

    r = _np.ceil(object_array.target_area_radius)

    figure = _plt.figure(figsize=_np.array([r, r]) * 2 / dpi,
                         dpi=dpi)
    if colours.background.colour is None:
        figure.patch.set_facecolor((0,0,0, 0))
    else:
        figure.patch.set_facecolor(colours.background.colour)
    axes = _plt.Axes(figure, [0., 0., 1, 1])
    axes.set_aspect('equal') # squared
    axes.set_axis_off()
    axes.set(xlim=[-1*r, r], ylim=[-1*r, r])
    figure.add_axes(axes)

    if colours.target_area.colour is not None:
        obj = _shapes.Dot(xy=(0, 0), diameter=object_array.target_area_radius * 2,
                          attribute=colours.target_area.colour)
        _draw_shape(axes, obj)

    if object_array.properties.numerosity > 0:
        if isinstance(object_array, _arrays.DotArray):
            # draw dots
            for xy, d, att in zip(object_array.xy,
                                  object_array.diameters,
                                  object_array.attributes):
                obj = _shapes.Dot(xy=xy, diameter=d)
                obj.attribute = _colour.Colour(att,
                                                    colours.default_object_colour)
                _draw_shape(axes, obj, opacity=colours.opacity_object)

        elif isinstance(object_array, _arrays.RectangleArray):
            # draw rectangle
            for xy, size, att in zip(object_array.xy,
                                     object_array.sizes,
                                     object_array.attributes):
                obj = _shapes.Rectangle(xy=xy, size=size)
                obj.attribute = _colour.Colour(att,
                                        colours.default_object_colour)
                _draw_shape(axes, obj, opacity=colours.opacity_object)

    # draw convex hulls
    if colours.field_area_positions.colour is not None and \
            object_array.properties.field_area_positions > 0:
        _draw_convex_hull(axes=axes,
                          points= object_array.properties.convex_hull_positions.xy,
                          convex_hull_colour=colours.field_area_positions.colour,
                          opacity=colours.opacity_guides)
    if colours.field_area.colour is not None and \
            object_array.properties.field_area > 0:
        _draw_convex_hull(axes=axes,
                          points=object_array.properties.convex_hull.xy,
                          convex_hull_colour=colours.field_area.colour,
                          opacity=colours.opacity_guides)
    #  and center of mass
    if colours.center_of_field_area.colour is not None:
        obj = _shapes.Dot(xy=object_array.center_of_field_area(),
                          diameter=10,
                          attribute=colours.center_of_field_area.colour)
        _draw_shape(axes, obj, opacity=colours.opacity_guides)
    if colours.center_of_mass.colour is not None:
        obj = _shapes.Dot(xy=object_array.center_of_mass(),
                          diameter=10,
                          attribute=colours.center_of_mass.colour)
        _draw_shape(axes, obj, opacity=colours.opacity_guides)

    return figure


def _draw_shape(axes, shape, opacity=1.0):
    assert isinstance(shape, (_shapes.Dot, _shapes.Rectangle))

    colour = _colour.Colour(shape.attribute)
    if isinstance(shape, _shapes.Dot):
        r = shape.diameter / 2
        plt_shape = _plt.Circle(xy=shape.xy, radius=r, color=colour.colour,
                                lw=0)
    elif isinstance(shape, _shapes.Rectangle):
        xy = (shape.left, shape.bottom)
        plt_shape = _plt.Rectangle(xy=xy,
                                   width=shape.width,
                                   height=shape.height,
                                   color=colour.colour, lw=0)

    else:
        raise NotImplementedError("Shape {} NOT YET IMPLEMENTED".format(type(shape)))

    plt_shape.set_alpha(opacity)
    axes.add_artist(plt_shape)

def _draw_convex_hull(axes, points, convex_hull_colour, opacity):
    # plot convey hull
    hull = _np.append(points, [points[0]], axis=0)
    for i in range(1, hull.shape[0]):
        line = _plt.Line2D(xdata=hull[i-1:i+1, 0],
                           ydata=hull[i-1:i+1, 1],
                           linewidth = 1, color = convex_hull_colour)
        line.set_alpha(opacity)
        axes.add_artist(line)




import random as _random
import numpy as _np
from math import log2 as _log2
from .._lib import geometry as _geometry
from .._lib.exceptions import NoSolutionError
from .. import arrays as _arrays
from ._props import VisualPropertyFlag as _flags

DEFAULT_SPACING_PRECISION = 0.0001
DEFAULT_ADAPT_FA2TA_RATIO = 0.5


def change_adapt_settings(default_spacing_precision=None,
                          default_adapt_fa2ta_ratio=None):
    """Changing class settings of property fitting.

    This changes the settings of the property fitting.


    Parameters
    ----------
    default_spacing_precision
    default_adapt_fa2ta_ratio

    Returns
    -------

    """
    global DEFAULT_ADAPT_FA2TA_RATIO
    global DEFAULT_SPACING_PRECISION
    if isinstance(default_spacing_precision, float):
        DEFAULT_SPACING_PRECISION = default_spacing_precision
    if isinstance(default_adapt_fa2ta_ratio, float):
        DEFAULT_ADAPT_FA2TA_RATIO = default_adapt_fa2ta_ratio

#FIXME coverage for all

def numerosity(object_array, value, center_of_field_area=False):
    TRY_OUT = 300
    _arrays._check_object_array(object_array)

    # make a copy for the deviant
    if value <= 0:
        object_array.clear()
    else:
        # add or remove random dots
        change_numerosity = value - object_array._properties.numerosity
        rnd = None
        for _ in range(abs(change_numerosity)):
            if center_of_field_area:
                ch = object_array._properties.convex_hull.indices
            else:
                ch = []
            for _ in range(TRY_OUT):
                rnd = _random.randint(0, object_array._properties.numerosity - 1)  # do not use np.random
                if rnd not in ch or change_numerosity > 0:
                    break

            if change_numerosity < 0:
                # remove dots
                object_array.delete(rnd)
            else:
                # add dot
                # copy a random dot
                rnd_object = next(object_array.iter_objects(rnd))
                try:
                    rnd_object = object_array.get_random_free_position(
                        ref_object=rnd_object, allow_overlapping=False,
                        inside_convex_hull=center_of_field_area
                    )
                except NoSolutionError:
                    # no free position
                    raise NoSolutionError("Can't make the deviant. No free position found.")
                object_array.add([rnd_object])

    return object_array


def average_diameter(dot_array, value):
    # changes diameter
    if not isinstance(dot_array, _arrays.DotArray):
        raise TypeError("Adapting diameter is not possible for {}.".format(
            type(dot_array).__name__))
    scale = value / dot_array.properties.average_dot_diameter
    dot_array._diameters = dot_array.diameters * scale
    dot_array.properties.reset()
    return dot_array

def average_rectangle_size(rect_array, value):
    # changes diameter
    if not isinstance(rect_array, _arrays.RectangleArray):
        raise TypeError("Adapting rectangle size is not possible for {}.".format(
            type(rect_array).__name__))
    try:
        width, height = value
    except TypeError:
        raise TypeError("Value ({}) has to tuple of 2 numerical (width, height).".format(
            value))

    scale = _np.array([width / rect_array.properties.average_rectangle_size[0],
                       height / rect_array.properties.average_rectangle_size[1]])
    rect_array._sizes = rect_array._sizes * scale
    rect_array.properties.reset()
    return rect_array

def total_surface_area(object_array, value):
    # changes diameter
    _arrays._check_object_array(object_array)
    a_scale = value / object_array.properties.total_surface_area
    if isinstance(object_array, _arrays.DotArray):
        object_array._diameters = _np.sqrt(
            object_array.surface_areas * a_scale) * 2 / _np.sqrt(
                    _np.pi)  # d=sqrt(4a/pi) = sqrt(a)*2/sqrt(pi)
    else: # rect
        object_array._sizes = object_array._sizes * _np.sqrt(a_scale)

    object_array.properties.reset()
    return object_array

def field_area(object_array, value, precision=None):
    """changes the convex hull area to a desired size with certain precision

    uses scaling radial positions if field area has to be increased
    uses replacement of outer points (and later re-scaling)

    iterative method can takes some time.
    """

    _arrays._check_object_array(object_array)
    if precision is None:
        precision = DEFAULT_SPACING_PRECISION

    if object_array.properties.field_area is _np.nan:
        return  object_array # not defined
    else:
        return _scale_field_area(object_array, value=value,
                                 precision=precision)

def _scale_field_area(object_array, value, precision):
    """change the convex hull area to a desired size by scale the polar
    positions  with certain precision

    iterative method can takes some time.

    Note: see doc string `_adapt_field_area`
    """
    _arrays._check_object_array(object_array)
    current = object_array.properties.field_area

    if current is None:
        return  # not defined

    scale = 1  # find good scale
    step = 0.1
    if value < current:  # current too larger
        step *= -1

    # centered points
    old_center = object_array.center_of_mass()
    object_array._xy = object_array.xy - old_center
    centered_polar = _geometry.cartesian2polar(object_array.xy)

    # iteratively determine scale
    while abs(current - value) > precision:

        scale += step

        object_array._xy = _geometry.polar2cartesian(centered_polar * [scale, 1])
        object_array.properties.reset()  # required at this point to recalc convex hull
        current = object_array.properties.field_area

        if (current < value and step < 0) or \
                (current > value and step > 0):
            step *= -0.2  # change direction and finer grain

    object_array._xy = object_array.xy + old_center
    object_array.properties.reset()
    return object_array

def coverage(object_array, value,
             precision=None,
             FA2TA_ratio=None):

    # FIXME check drifting outwards if extra space is small and adapt_FA2TA_ratio=1
    # FIXME when to realign, realignment changes field_area!
    """this function changes the area and remixes to get a desired density
    precision in percent between 1 < 0

    ratio_area_convex_hull_adaptation:
        ratio of adaptation via area or via convex_hull (between 0 and 1)

    """
    _arrays._check_object_array(object_array)

    print("WARNING: _adapt_coverage is a experimental ")
    # dens = convex_hull_area / total_surface_area
    if FA2TA_ratio is None:
        FA2TA_ratio = DEFAULT_ADAPT_FA2TA_RATIO
    elif FA2TA_ratio < 0 or FA2TA_ratio > 1:
        FA2TA_ratio = 0.5
    if precision is None:
        precision = DEFAULT_SPACING_PRECISION

    total_area_change100 = (value * object_array.properties.field_area) - \
                           object_array.properties.total_surface_area
    d_change_total_area = total_area_change100 * (1 - FA2TA_ratio)
    if abs(d_change_total_area) > 0:
        total_surface_area(object_array.properties.total_surface_area + \
                           d_change_total_area)

    return field_area(object_array.properties.total_surface_area / value,
               precision=precision)

def average_perimeter(object_array, value):
    _arrays._check_object_array(object_array)
    total_peri = value * object_array.properties.numerosity
    return total_perimeter(object_array, total_peri)

def total_perimeter(object_array, value):
    if isinstance(object_array, _arrays.DotArray):
        tmp = value / (object_array.properties.numerosity * _np.pi)
        return average_diameter(object_array, tmp)
    elif isinstance(object_array, _arrays.RectangleArray):
        scale = value / object_array.properties.total_perimeter
        new_size = object_array.properties.average_rectangle_size * scale
        return average_rectangle_size(object_array, new_size)
    else:
        _arrays._check_object_array(object_array)

def average_surface_area(object_array, value):
    _arrays._check_object_array(object_array)
    ta = object_array.properties.numerosity * value
    return total_surface_area(object_array, ta)

def log_spacing(object_array, value, precision=None):
    _arrays._check_object_array(object_array)
    logfa = 0.5 * value + 0.5 * _log2(
        object_array.properties.numerosity)
    return field_area(object_array, value=2 ** logfa, precision=precision)

def log_size(object_array, value):
    _arrays._check_object_array(object_array)
    logtsa = 0.5 * value + 0.5 * _log2(object_array.properties.numerosity)
    return total_surface_area(object_array, 2 ** logtsa)

def sparcity(object_array, value, precision=None):
    _arrays._check_object_array(object_array)
    return field_area(object_array, value=value * object_array.properties.numerosity,
                      precision=precision)

def visual_property(object_array, property_flag, value):
    """
    adapt_properties: continuous property or list of continuous properties
    several properties to be adapted
    if adapt dot array is specified, array will be adapt to adapt_dot_array, otherwise
    the values defined in adapt_properties is used.
    some adapting requires realignement to avoid overlaps. However,
    realigment might result in a different field area. Thus, realign after
    adapting for  Size parameter and realign before adapting space
    parameter.

    """

    # type check
    if not isinstance(property_flag, _flags):
        raise ValueError("{} is not a visual feature.".format(property_flag))

    # Adapt
    if property_flag == _flags.AV_DOT_DIAMETER:
        return average_diameter(object_array, value=value)

    elif property_flag == _flags.NUMEROSITY:
        return numerosity(object_array, value=value)

    elif property_flag == _flags.AV_PERIMETER:
        return average_perimeter(object_array, value=value)

    elif property_flag == _flags.TOTAL_PERIMETER:
        return total_perimeter(object_array, value=value)

    elif property_flag == _flags.AV_SURFACE_AREA:
        return average_surface_area(object_array, value=value)

    elif property_flag == _flags.TOTAL_SURFACE_AREA:
        return total_surface_area(object_array, value=value)

    elif property_flag == _flags.LOG_SIZE:
        return log_size(object_array, value=value)

    elif property_flag == _flags.LOG_SPACING:
        return log_spacing(object_array, value=value)

    elif property_flag == _flags.SPARSITY:
        return sparcity(object_array, value=value)

    elif property_flag == _flags.FIELD_AREA:
        return field_area(object_array, value=value)

    elif property_flag == _flags.COVERAGE:
        return coverage(object_array, value=value)

    else:
        raise NotImplementedError("Not implemented for {}".format(
            property_flag.label()))


"""
NonSyNum package

Creating Non-Symbolic Number Displays
"""

__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'
__version__ = '0.11.5-3'

from sys import version_info as _python_version_info
if not(_python_version_info[0] >= 3 and _python_version_info[1] >= 6):
    raise RuntimeError("PyNsN {0} ".format(__version__) +
                      "is not compatible with Python {0}.{1}. ".format(
                                                    _python_version_info[0],
                                                    _python_version_info[1]) +
                      "Please use Python 3.6 or later.")

from . import shapes
from . import arrays
from . import visual_properties
from . import random_array
from . import distributions
from ._lib.exceptions import NoSolutionError




__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

from PIL import Image as _Image
from PIL import ImageDraw as _ImageDraw
import numpy as _np
from . import _colour
from .._lib.geometry import cartesian2image_coordinates as _c2i_coord
from .. import shapes as _shapes
from .. import arrays as _arrays
from .._lib import misc as _misc


from ._colour import ImageColours # make available

#FIXME pillow supports no alpha

def create(object_array, colours=None, antialiasing=True, _gabor_filter=None):
    # ImageParameter
    """use PIL colours (see PIL.ImageColor.colormap)

    returns pil image

    antialiasing: Ture or integer

    gabor_filter: from PIL.ImageFilter
    default_dot_colour: if colour is undefined in _lib
    """

    _arrays._check_object_array(object_array)
    if colours is None:
        colours = _colour.ImageColours()
    if not isinstance(colours, _colour.ImageColours):
        raise TypeError("Colours must be of type image.ImageColours")

    if isinstance(antialiasing, bool):
        if antialiasing:  # (not if 1)
            aaf = 2  # AA default
        else:
            aaf = 1
    else:
        try:
            aaf = int(antialiasing)
        except ValueError:
            aaf = 1

    # prepare the pil image, make target area if required
    image_size = int(_np.ceil(object_array.target_area_radius) * 2) * aaf
    img = _Image.new("RGBA", (image_size, image_size),
                     color=colours.background.colour)

    if colours.target_area.colour is not None:
        obj = _shapes.Dot(xy=_c2i_coord(_np.zeros(2), image_size),
                          diameter=image_size,
                          attribute=colours.target_area.colour)
        _draw_shape(img, obj)

    if object_array.properties.numerosity > 0:
        image_coord = _c2i_coord(object_array.xy * aaf, image_size)
        if isinstance(object_array, _arrays.DotArray):
            # draw dots
            for xy, d, att in zip(image_coord, object_array.diameters * aaf,
                                  object_array.attributes):
                obj = _shapes.Dot(xy=xy, diameter=d)
                obj.attribute = _colour.Colour(att,
                                               colours.default_object_colour)
                _draw_shape(img, obj)

        elif isinstance(object_array, _arrays.RectangleArray):
            # draw rectangle
            for xy, size, att in zip(image_coord,
                                     object_array.sizes * aaf,
                                     object_array.attributes):
                obj = _shapes.Rectangle(xy=xy, size=size,
                                        attribute=att)
                if obj.get_picture() is None:
                    # rect
                    obj.attribute = _colour.Colour(att,
                                colours.default_object_colour)
                _draw_shape(img, obj)

        # draw convex hulls
        if colours.field_area_positions.colour is not None and \
                object_array.properties.field_area_positions > 0:
            _draw_convex_hull(img=img,
                              points=_c2i_coord(
                                  object_array.properties.convex_hull_positions.xy * aaf, image_size),
                              convex_hull_colour=colours.field_area_positions.colour)
        if colours.field_area.colour is not None and \
                object_array.properties.field_area > 0:
            _draw_convex_hull(img=img,
                              points=_c2i_coord(
                                  object_array.properties.convex_hull.xy * aaf,
                                  image_size),
                              convex_hull_colour=colours.field_area.colour)
        #  and center of mass
        if colours.center_of_field_area.colour is not None:
            obj = _shapes.Dot(xy=_c2i_coord(object_array.center_of_field_area() * aaf, image_size),
                              diameter=10 * aaf,
                              attribute=colours.center_of_field_area.colour)
            _draw_shape(img, obj)
        if colours.center_of_mass.colour is not None:
            obj = _shapes.Dot(xy=_c2i_coord(object_array.center_of_mass() * aaf, image_size),
                              diameter=10 * aaf,
                              attribute=colours.center_of_mass.colour)
            _draw_shape(img, obj)

    # rescale for antialiasing
    if aaf != 1:
        image_size = int(image_size / aaf)
        img = img.resize((image_size, image_size), _Image.LANCZOS)

    # FIXME gabor needed?
    if _gabor_filter is not None:
        try:
            img = img.filter(_gabor_filter)
        except:
            raise RuntimeError("Can't apply gabor_filter {}".format(_gabor_filter))

    return img


def _draw_shape(img, shape):
    # draw object
    colour = _colour.Colour(shape.attribute)
    if isinstance(shape, _shapes.Dot):
        r = shape.diameter / 2
        _ImageDraw.Draw(img).ellipse((shape.x - r, shape.y - r,
                                      shape.x + r, shape.y + r),
                                     fill=colour.colour)
    elif isinstance(shape, _shapes.Rectangle):
        if shape.get_picture() is not None:
            # picture
            shape_size = (round(shape.width), round(shape.height))
            target_box = (round(shape.left), round(shape.bottom),
                          round(shape.right), round(shape.top)) # reversed y axes
            pict = _Image.open(shape.get_picture(), "r")
            if pict.size[0] != shape_size[0] or pict.size[1] != shape_size[1]:
                pict = pict.resize(shape_size, resample=_Image.ANTIALIAS)

            tr_layer = _Image.new('RGBA', img.size, (0, 0, 0, 0))
            tr_layer.paste(pict, target_box)
            res = _Image.alpha_composite(img, tr_layer)
            img.paste(res)
        else:
            # rectangle shape
            _ImageDraw.Draw(img).rectangle((shape.left, shape.top,
                                        shape.right, shape.bottom),
                                       fill=colour.colour) # FIXME decentral shapes seems to be bit larger than with pyplot

    else:
        raise NotImplementedError("Shape {} NOT YET IMPLEMENTED".format(type(shape)))


def _draw_convex_hull(img, points, convex_hull_colour):
    # plot convey hull

    last = None
    draw = _ImageDraw.Draw(img)
    for p in _np.append(points, [points[0]], axis=0):
        if last is not None:
            draw.line(_np.append(last, p).tolist(),
                      width=2,
                      fill=convex_hull_colour)
        last = p




from ._base_array import BaseArray
from ._dot_array import DotArray
from ._rect_array import RectangleArray

# helper for type checking and error raising error
def _check_base_array(obj):
    if not isinstance(obj, BaseArray):
        raise TypeError("DotArray, RectangleArray or AttributeArray expected, but not {}".format(
            type(obj).__name__))

def _check_object_array(obj):
    if not isinstance(obj, (DotArray, RectangleArray)):
        raise TypeError("DotArray or RectangleArray expected, but not {}".format(
            type(obj).__name__))




import random
import numpy as np

from .._lib import  geometry


def _jitter_identical_positions(xy, jitter_size=0.1):
    """jitters points with identical position"""

    for idx, ref_object in enumerate(xy):
        identical = np.where(np.all(np.equal(xy, ref_object), axis=1))[0]  # find identical positions
        if len(identical) > 1:
            for x in identical:  # jitter all identical positions
                if x != idx:
                    xy[x, :] = xy[x, :] - geometry.polar2cartesian(
                        [[jitter_size, random.random() * 2 * np.pi]])[0]

    return xy

def radial_replacement_from_reference_dots(xy, ref_pos_id,
                                            neighbour_ids, replacement_size):
    """remove neighbouring position radially from reference position
    helper function, typically used for realign
    """

    # check if there is an identical position and jitter to avoid fully overlapping positions
    if np.sum(np.all(xy[neighbour_ids,] == xy[ref_pos_id, :],
                   axis=1)) > 0:
        xy = _jitter_identical_positions(xy)

    # relative polar positions to reference_dot
    tmp_polar = geometry.cartesian2polar(xy[neighbour_ids, :] - xy[ref_pos_id, :])
    tmp_polar[:, 0] = 0.000000001 + replacement_size # determine movement size
    tmp_xy = geometry.polar2cartesian(tmp_polar)
    xy[neighbour_ids, :] = np.array([xy[neighbour_ids, 0] + tmp_xy[:, 0],
                                           xy[neighbour_ids, 1] + tmp_xy[:, 1]]).T
    return xy

def remove_overlap_from_inner_to_outer(xy, min_dist_between, distance_matrix_function):
    """returns xy and boolean, if replacements were required"""
    assert callable(distance_matrix_function)

    replacement_required = False
    # from inner to outer remove overlaps
    for i in np.argsort(geometry.cartesian2polar(xy, radii_only=True)):
        dist_mtx = distance_matrix_function(between_positions=False)
        dist = dist_mtx[i,:]
        idx_overlaps = np.where(dist < min_dist_between)[0].tolist()  # overlapping dot ids
        if len(idx_overlaps) > 1:
            replacement_required = True
            idx_overlaps.remove(i)  # don't move yourself
            replace_size =min_dist_between - dist[idx_overlaps]  # dist is mostly negative, because of overlap
            xy = radial_replacement_from_reference_dots(
                                         xy=xy,
                                         ref_pos_id=i,
                                         neighbour_ids=idx_overlaps,
                                         replacement_size=replace_size)
    return replacement_required



"""
Object Cloud
"""

__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

from hashlib import md5
import json
from copy import deepcopy
import random
from abc import ABCMeta, abstractmethod

import numpy as np
from scipy import spatial
from .._lib import misc, geometry
from .._lib.exceptions import NoSolutionError
from ..visual_properties._props import ArrayProperties
from ..visual_properties import fit
from .. import shapes


class BaseArray(object):
    """Class for attributes on two dimensional space"""

    def __init__(self, target_area_radius,
                 min_dist_between=2,
                 min_dist_area_boarder=1,
                 xy=None,
                 attributes=None):
        """Numpy Position lists with attributes for optimized for numpy calculations

        Abstract class for implementation of dot and rect
        """
        self.target_area_radius = target_area_radius
        self.min_dist_between = min_dist_between
        self.min_dist_area_boarder = min_dist_area_boarder

        self._xy = np.array([])
        self._attributes = np.array([])
        self._properties = ArrayProperties(self)

        if xy is not None:
            self._append_xy_attribute(xy=xy, attributes=attributes)

    def _append_xy_attribute(self, xy, attributes=None):
        """returns number of added rows"""
        xy = misc.numpy_array_2d(xy)
        if not isinstance(attributes, (tuple, list)):
            attributes = [attributes] * xy.shape[0]

        if len(attributes) != xy.shape[0]:
            raise ValueError("Bad shaped data: attributes have not " +
                             "the same length as the coordinates")

        self._attributes = np.append(self._attributes, attributes)
        if len(self._xy) == 0:
            empty = np.array([]).reshape((0, 2))  # ensure good shape of self.xy
            self._xy = np.append(empty, xy, axis=0)
        else:
            self._xy = np.append(self._xy, xy, axis=0)
        self._properties.reset()
        return xy.shape[0]

    def __str__(self):
        prop_text = self._properties.as_text(extended_format=True)
        rtn = "- {}".format(type(self).__name__)
        rtn += "\n " + prop_text[1:]  # replace "-" with " "
        return rtn

    @property
    def xy(self):
        return self._xy

    @property
    def xy_rounded_integer(self):
        """rounded to integer"""
        return np.array(np.round(self._xy))

    @property
    def attributes(self):
        return self._attributes

    @property
    def properties(self):
        return self._properties

    @property
    def surface_areas(self):
        """per definition always zero"""
        return np.array([0] * len(self._xy))

    @property
    def perimeter(self):
        """per definition always zero"""
        return np.array([0] * len(self._xy))

    def set_attributes(self, attributes):
        """Set all attributes

        Parameter
        ---------
        attributes:  attribute (string) or list of attributes

        """

        if isinstance(attributes, (list, tuple)):
            if len(attributes) != self._properties.numerosity:
                raise ValueError("Length of attribute list does not adapt the " + \
                                 "size of the dot array.")
            self._attributes = np.array(attributes)
        else:
            self._attributes = np.array([attributes] * self._properties.numerosity)

    @property
    def hash(self):
        """md5_hash of positions and perimeter"""
        m = md5()
        m.update(
            self._xy.tobytes())  # to_byte required: https://stackoverflow.com/questions/16589791/most-efficient-property-to-hash-for-numpy-array
        try:
            m.update(self.perimeter.tobytes())
        except AttributeError:
            pass
        m.update(self._attributes.tobytes())
        return m.hexdigest()

    def center_of_positions(self):
        """Center of all object positions
        Notes
        -----
        if you want centre that takes into account the object size, use
        center_of_field_area
        """
        return geometry.center_of_positions(self._xy)

    def clear(self):
        self._xy = np.array([])
        self._attributes = np.array([])
        self._properties.reset()

    def delete(self, index):
        self._xy = np.delete(self._xy, index, axis=0)
        self._attributes = np.delete(self._attributes, index)
        self._properties.reset()

    def copy(self, indices=None, deepcopy=True):
        """returns a (deep) copy of the dot array.

        It allows to copy a subset of dot only.

        """

        if len(self._xy) == 0:
            return BaseArray(target_area_radius=self.target_area_radius,
                             min_dist_between=self.min_dist_between,
                             min_dist_area_boarder=self.min_dist_area_boarder)

        if indices is None:
            indices = list(range(len(self._xy)))

        if deepcopy:
            return BaseArray(target_area_radius=self.target_area_radius,
                             min_dist_between=self.min_dist_between,
                             min_dist_area_boarder=self.min_dist_area_boarder,
                             xy=self._xy[indices, :].copy(),
                             attributes=self._attributes[indices].copy())
        else:
            return BaseArray(target_area_radius=self.target_area_radius,
                             min_dist_between=self.min_dist_between,
                             min_dist_area_boarder=self.min_dist_area_boarder,
                             xy=self._xy[indices, :],
                             attributes=self._attributes[indices])

    def as_dict(self):
        """
        """
        d = {"type": type(self).__name__,
             "target_area_radius": self.target_area_radius,
             "min_dist_between": self.min_dist_between,
             "min_dist_area_boarder": self.min_dist_area_boarder}
        d.update({"xy": self._xy.tolist()})
        if len(self._attributes) > 0 and misc.is_all_equal(self._attributes):
            d.update({"attributes": self._attributes[0]})
        else:
            d.update({"attributes": self._attributes.tolist()})
        return d

    def read_from_dict(self, dict):
        """read dot array from dict"""
        self.target_area_radius = dict["target_area_radius"]
        self.min_dist_between = dict["min_dist_between"]
        self.min_dist_area_boarder = dict["min_dist_area_boarder"]
        self._xy = np.array(dict["xy"])
        if not isinstance(dict["attributes"], (list, tuple)):
            att = [dict["attributes"]] * self._properties.numerosity
        else:
            att = dict["attributes"]
        self._attributes = np.array(att)
        self._properties.reset()

    def json(self, indent=None, include_hash=False):
        """"""
        # override and extend as_dict not this function

        d = self.as_dict()
        if include_hash:
            d.update({"hash": self.hash})
        if not indent:
            indent = None
        return json.dumps(d, indent=indent)

    def save(self, json_file_name, indent=None, include_hash=False):
        """"""
        with open(json_file_name, 'w') as fl:
            fl.write(self.json(indent=indent, include_hash=include_hash))

    def load(self, json_file_name):
        # override and extend read_from_dict not this function
        with open(json_file_name, 'r') as fl:
            dict = json.load(fl)
        self.read_from_dict(dict)


class ABCObjectArray(BaseArray, metaclass=ABCMeta):

    @property
    @abstractmethod
    def surface_areas(self):
        pass

    @property
    @abstractmethod
    def perimeter(self):
        pass

    @abstractmethod
    def as_dict(self):
        return super().as_dict()

    @abstractmethod
    def read_from_dict(self, dict):
        return super().read_from_dict()

    @abstractmethod
    def copy(self, indices=None, deepcopy=True):
        """returns a (deep) copy of the dot array.

        It allows to copy a subset of dot only.

        """
        pass

    @abstractmethod
    def iter_objects(self, indices=None):
        pass

    @abstractmethod
    def add(self, something):
        pass

    @abstractmethod
    def find(self, attribute):
        pass

    @abstractmethod
    def distances(self, ref_object):
        # override this method
        pass

    @abstractmethod
    def check_stand_outs(self):
        pass

    def join(self, object_array):
        """add another object arrays"""
        self.add(object_array.iter_objects())

    def center_of_field_area(self):
        return geometry.center_of_positions(self.properties.convex_hull.xy)

    def distances_matrix(self, between_positions=False):
        """between position ignores the dot size"""
        if between_positions:
            return spatial.distance.cdist(self._xy, self._xy)
        # matrix with all distance between all points
        dist = np.array([self.distances(d) for d in self.iter_objects()])
        return dist

    def check_overlaps(self):
        """return pairs of indices of overlapping of objects
        numpy.array
        """
        dist = self.distances_matrix(between_positions=False)
        overlap = np.where(np.triu(dist, k=1) < 0)
        return np.array(overlap).T

    def center_of_mass(self):
        weighted_sum = np.sum(self._xy * self.perimeter[:, np.newaxis], axis=0)
        return weighted_sum / np.sum(self.perimeter)

    def get_random_free_position(self, ref_object,
                                 allow_overlapping=False,
                                 inside_convex_hull=False,
                                 occupied_space=None):
        """returns the copy of object of at a random free position

        raise exception if not found
        occupied space: see generator generate
        """

        N_ATTEMPTS = 3000

        if isinstance(ref_object, shapes.Dot):
            object_size = ref_object.diameter / 2.0
        elif isinstance(ref_object, shapes.Rectangle):
            object_size = max(ref_object.size)
        else:
            raise NotImplementedError("Not implemented for {}".format(
                type(ref_object).__name__))
        if occupied_space is not None and \
                not isinstance(occupied_space, ABCObjectArray): #FIXME check
            raise TypeError("Occupied_space has to be a Dot or Rectangle Array or None.")

        area_rad = self.target_area_radius - self.min_dist_area_boarder - object_size
        rtn_object = deepcopy(ref_object)  # tested deepcopy required

        cnt = 0
        while True:
            cnt += 1
            ##  polar method seems to produce central clustering
            #  proposal_polar =  np.array([random.random(), random.random()]) *
            #                      (target_radius, TWO_PI)
            rtn_object.xy = np.array([random.random(), random.random()]) \
                            * 2 * area_rad - area_rad

            # is outside area
            if isinstance(ref_object, shapes.Dot):
                bad_position = area_rad <= rtn_object.polar_radius
            else:
                # Rect: check if one edge is outside
                bad_position = False
                for e in rtn_object.iter_edges():
                    if e.polar_radius >= area_rad:
                        bad_position = True
                        break

            if not bad_position and not allow_overlapping:
                # find bad_positions
                dist = self.distances(rtn_object)
                if isinstance(occupied_space, ABCObjectArray):
                    dist = np.append(dist, occupied_space.distances(rtn_object))
                bad_position = sum(dist < self.min_dist_between) > 0  # at least one is overlapping

            if not bad_position and inside_convex_hull:
                # use only those that do not change the convex hull
                tmp_array = self.copy(deepcopy=True)
                tmp_array.add([rtn_object])
                bad_position = tmp_array.properties.convex_hull != \
                               self.properties.convex_hull

            if not bad_position:
                return rtn_object
            elif cnt > N_ATTEMPTS:
                raise NoSolutionError(u"Can't find a free position")

    def shuffle_all_positions(self, allow_overlapping=False):
        """might raise an exception"""
        # find new position for each dot
        # mixes always all position (ignores dot limitation)

        all_objects = list(self.iter_objects())
        self.clear()
        for obj in all_objects:
            try:
                new = self.get_random_free_position(obj,
                                        allow_overlapping=allow_overlapping)
            except NoSolutionError as e:
                raise NoSolutionError("Can't shuffle dot array. No free positions found.")
            self.add([new])


    def get_number_deviant(self, change_numerosity, preserve_field_area=False):
        """number deviant
        """
        object_array = self.copy()
        new_num = self.properties.numerosity + change_numerosity
        fit.numerosity(object_array, value=new_num,
                       center_of_field_area=preserve_field_area)
        return object_array

    def replace_overlapping_objects(self, center_of_field_area=False,
                                    lenient=True):
        """
        Returns
        Parameters
        ----------
        center_of_field_area
        lenient

        Returns
        -------
        convex_hull_had_changed
        """

        warning_info = "Field area had to change, because two overlapping " + \
                       "objects on convex hull"
        convex_hull_had_changed = False

        overlaps = self.check_overlaps()
        while len(overlaps):
            if center_of_field_area:
                # check if overlaps are on convex hull
                # do not replace convexhull objects and try to
                # take that one not on convex hull or raise error/warning
                ch_idx = self.properties.convex_hull.indices
                if overlaps[0, 0] not in ch_idx:
                    idx = overlaps[0, 0]
                elif overlaps[0, 1] not in ch_idx:
                    idx = overlaps[0, 1]
                elif lenient:
                    # warning
                    convex_hull_had_changed = True
                    idx = overlaps[0, 0]
                else:
                    raise NoSolutionError("Can't replace overlap and keep convex hull unchanged. " +
                                          warning_info)
            else:
                idx = overlaps[0, 0]

            obj = next(self.iter_objects(idx))
            self.delete(idx)
            obj = self.get_random_free_position(ref_object=obj,
                                                inside_convex_hull=center_of_field_area)
            self.add([obj])
            overlaps = self.check_overlaps()

        if convex_hull_had_changed:
            print("Warning: " + warning_info)

        return convex_hull_had_changed

    @abstractmethod
    def center_array(self):
        """places array in target area as central and possible and tries to
        remove any stand_outs"""
        pass

    @abstractmethod
    def realign(self):
        raise NotImplementedError()

    def get_split_arrays(self):
        """returns a list of arrays
        each array contains all dots of with particular colour"""
        att = self._attributes
        att[np.where(att == None)] = "None"  # TODO check "is none"

        rtn = []
        for c in np.unique(att):
            if c is not None:
                da = self.copy(indices=0, deepcopy=False)  # fast. shallow copy with just one object
                da.clear()
                da.add(self.find(attribute=c))
                rtn.append(da)
        return rtn


# TODO  everywhere: file header doc and author information



COL_TARGET_AREA = "#FFF0D9"
COL_FIELD_AREA_POSITIONS = None
COL_FIELD_AREA = None
COL_CENTER_OF_FIELD_AREA = None
COL_CENTER_OF_MASS = None
COL_BACKGROUND = None
COL_DEFAULT_OBJECT = "darkgreen"
OPACITY_OBJECT = 1
OPACITY_GUIDES = 0.5


from .._lib.misc import dict_to_text
from .. import shapes
from ..distributions import PyNSNDistribution, Constant, _round_samples


class SizeDistribution(object):

    def __init__(self, diameter=None, width=None,
                 height=None, rectangle_proportion=None):

        rect_parameter_set = [width is not None,
                              height is not None,
                              rectangle_proportion is not None]

        is_rect = sum(rect_parameter_set)>0
        is_dot = diameter is not None
        if is_dot and is_rect:
         raise TypeError("Please define either dot or rectangle parameter, not both.")
        elif not is_dot and not is_rect:
         raise TypeError("No size distribution define. Please define either diameter or width and height.")
        elif is_rect and sum(rect_parameter_set) != 2:
            raise TypeError("Define rectangle width and height or, alternatively, rectangle proportion together with either width or height.")

        if isinstance(diameter, (int, float)):
            diameter = Constant(diameter)
        if isinstance(width, (int, float)):
            width = Constant(width)
        if isinstance(height, (int, float)):
            height = Constant(height)
        if isinstance(rectangle_proportion, (int, float)):
            rectangle_proportion = Constant(rectangle_proportion)

        err = "{} has to be a PyNSNDistribution or single number (constant)"
        if diameter is not None and not isinstance(diameter, PyNSNDistribution):
            raise TypeError(err.format("dot_diameter"))
        if width is not None and not isinstance(width, PyNSNDistribution):
            raise TypeError(err.format("rectangle_width"))
        if height is not None and not isinstance(height, PyNSNDistribution):
            raise TypeError(err.format("rectangle_height"))
        if rectangle_proportion is not None and \
                not isinstance(rectangle_proportion, PyNSNDistribution):
            raise TypeError(err.format("rectangle_proportion"))

        self.diameter = diameter
        self.width = width
        self.height = height
        self.proportion = rectangle_proportion

    def sample(self, n, round_to_decimals=None):
        """return list objects (Dot or Rect) with random size
        all positions = (0,0)
        """
        if self.diameter is not None:
            diameter = self.diameter.sample(n)
            return [shapes.Dot(xy=(0, 0), diameter=dia) for dia in diameter]
        else:
            #Rect
            try:
                width = self.width.sample(n)
            except AttributeError:
                width = None
            try:
                height = self.height.sample(n)
            except AttributeError:
                height = None
            try:
                proportion = self.proportion.sample(n)
            except AttributeError:
                proportion = None

            if height is None:
                height = width * proportion
            elif width is None:
                width = height / proportion

            if round_to_decimals is not None:
                width = _round_samples(width, round_to_decimals=round_to_decimals)
                height = _round_samples(width, round_to_decimals=round_to_decimals)

            return [shapes.Rectangle(xy=(0, 0), size=s) for s in zip(width, height)]


    def as_dict(self):
        rtn = {}
        try:
            rtn.update({"diameter": self.diameter.as_dict()})
        except AttributeError:
            pass
        try:
            rtn.update({"width": self.width.as_dict()})
        except AttributeError:
            pass
        try:
            rtn.update({"height": self.height.as_dict()})
        except AttributeError:
            pass
        try:
            rtn.update({"proportion": self.proportion.as_dict()})
        except AttributeError:
            pass
        return rtn


    def __str__(self):
        return dict_to_text(self.as_dict(), col_a=12, col_b=7)


